#include <Application.h>
#include <Bitmap.h>
#include <Font.h>
#include <View.h>
#include <Point.h>
#include <Screen.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#define __BEOSSPRITE_CPP__
#include "beossprite.h"
#include "beosmain.h"
#include "config.h"


void
W_Initialize()
{
    // Get fonts.
    struct font_height fh;

    BFont *normal = new BFont(be_fixed_font);
    normal->SetSize(10.0);
    normal->SetFlags(B_DISABLE_ANTIALIASING);
    normal->GetHeight(&fh);
    W_RegularFont = (W_Font)normal;
    W_Textwidth = int(0.999 + normal->StringWidth("W"));
    W_Textheight = 1 + int(fh.ascent + 0.999) + int(fh.descent + 0.999) +
	int(fh.leading + 0.999);

    BFont *big = new BFont(be_fixed_font);
    big->SetSize(34.0);
    normal->SetFlags(B_DISABLE_ANTIALIASING);
    big->GetHeight(&fh);
    W_BigFont = (W_Font)big;
    W_BigTextwidth = int(0.999 + big->StringWidth("W"));
    W_BigTextheight = int(fh.ascent + 0.999) + int(fh.descent + 0.999) +
	int(fh.leading + 0.999);
}


void
W_CacheClearArea(W_Window window, int x, int y, int width, int height)
{
    W_ClearArea(window, x, y, width, height);
}


void
W_ClearArea(W_Window window, int x, int y, unsigned int width, unsigned int height)
{
    BView *buffer = (BView *)window->buffer;
    buffer->Window()->Lock();
    buffer->SetHighColor(W_Black);
    buffer->FillRect(BRect(x, y, x + width-1, y + height-1));
    buffer->Window()->Unlock();
}


void
W_ClearWindow(W_Window window)
{
    BView *buffer = (BView *)window->buffer;
    buffer->Window()->Lock();
    buffer->SetHighColor(W_Black);
    buffer->FillRect(buffer->Bounds());
    buffer->Window()->Unlock();
}


void
W_DisplayBuffer(W_Window window)
{
    BView *buffer = (BView *)window->buffer;
    buffer->Window()->Lock();
    buffer->Window()->Sync();
    buffer->Window()->Unlock();

    BView *view = (BView *)window->view;

    BScreen screen(view->Window());
    screen.WaitForRetrace();

    BBitmap *bitmap = (BBitmap *)window->bitmap;
    view->Window()->Lock();
    view->SetDrawingMode(B_OP_COPY);
    view->DrawBitmap(bitmap, BPoint(0, 0));
    view->Window()->Flush();
    view->Window()->Unlock();
}


void
W_DrawImage(W_Window window, int x, int y, int frame, W_Image *image, W_Color color)
{
    int height, width;

    if (!image->loaded) {
	W_LoadImage(image);
    }

    if (frame < 0) {	// draw the whole thing regardless of frames
	height = image->height * image->frames;
	frame = 0;
    }
    else {		// draw the frame given
	height = image->height;
	frame = frame % image->frames;
    }
    width = image->width;

    BRect src(0, frame * height, width-1, (frame+1) * height - 1);
    BRect dest(x, y, x + width-1, y + height-1);
    BView *buffer = (BView *)window->buffer;
    buffer->Window()->Lock();
    buffer->SetDrawingMode(B_OP_OVER);
    buffer->DrawBitmapAsync((BBitmap *)image->pixmap, src, dest);
    buffer->Window()->Unlock();
}


void
W_DrawPoint(W_Window window, int x, int y, W_Color color)
{
    BView *buffer = (BView *)window->buffer;

    buffer->Window()->Lock();
    buffer->SetHighColor(color);
    buffer->SetDrawingMode(B_OP_COPY);
    buffer->StrokeLine(BPoint(x, y), BPoint(x, y));
    buffer->Window()->Unlock();
}


int
W_EventsPending(W_Window window)
{
    GalagaView *view = (GalagaView *)window->view;
    return view->eventsPending();
}


void
W_Flush(W_Window window)
{
}


void
W_FlushClearAreaCache(W_Window window)
{
}


void
W_GetMouse(W_Window window, int *x, int *y, int *but)
{
    BView *view = (BView *)window->view;
    BPoint cursor;
    uint32 buttons;
    view->Window()->Lock();
    view->GetMouse(&cursor, &buttons, false);
    view->Window()->Unlock();
    *x = (int)cursor.x;
    *y = (int)cursor.y;
    *but = (buttons == 0) ? 0 : W_LBUTTON;
}


void
W_GrabPointer(W_Window window)
{
    BView *view = (BView *)window->view;
    view->Window()->Lock();
    view->SetEventMask(B_POINTER_EVENTS);
    view->Window()->Unlock();
}


int
W_LoadImage(W_Image *image)
{
    extern char *app_dir;
    char filename[512];

    sprintf(filename, "%s/%s/%s%s", app_dir, IMAGEDIR, image->filename, ".bin");
    FILE *f = fopen(filename, "r");
    if (f == NULL) {
	printf("Can't read image file %s\n", filename);
	return 1;
    }

    // Get image parameters.
    int width, height, nframes;

    fread((void*)&width, sizeof(int), 1, f);
    fread((void*)&height, sizeof(int), 1, f);
    fread((void*)&nframes, sizeof(int), 1, f);

    width = B_LENDIAN_TO_HOST_INT32(width);
    height = B_LENDIAN_TO_HOST_INT32(height);
    nframes = B_LENDIAN_TO_HOST_INT32(nframes);

    // Get image rgb_color palette.
    int ncolors;
    fread((void*)&ncolors, sizeof(int), 1, f);
    ncolors = B_LENDIAN_TO_HOST_INT32(ncolors);
    rgb_color *palette = new rgb_color[ncolors];
    fread((void*)palette, sizeof(rgb_color), ncolors, f);
    // rgb_color is endian-independent, since it contains individual bytes.
    for (int i = 0; i < ncolors; i++) {
	if (palette[i].alpha == 0) {
	    palette[i] = B_TRANSPARENT_32_BIT;
	}
    }

    // Get pixels.  Each byte is an index into the palette.
    uint8 *pixels = new uint8[width * height];
    fread((void*)pixels, 1, width * height, f);
    fclose(f);

    // Build BBitmap object.
    BBitmap *bitmap = new BBitmap(BRect(0, 0, width-1, height-1), B_CMAP8);

    // Rows within the BBitmap object begin on 4-byte aligned offsets!
    BScreen screen;
    uint8 *bits = (uint8*)bitmap->Bits();
    uint8 *row = pixels;
    int offset = 0;
    for (int i = 0; i < height; i++) {
	// Translate this row of pixels into system colors.
	for (int col = 0; col < width; col++) {
	    bits[offset + col] = screen.IndexForColor(palette[row[col]]);
	}
	offset += bitmap->BytesPerRow();
	row += width;
    }

    delete pixels;
    delete palette;

    image->width = width;
    image->height = height / nframes;
    image->frames = nframes;
    image->pixmap = (unsigned char *)bitmap;
    image->loaded = 1;

    return 0;
}


void
W_MakeLine(W_Window window, int X0, int Y0, int X1, int Y1, W_Color color)
{
    BView *buffer = (BView *)window->buffer;

    buffer->Window()->Lock();
    buffer->SetHighColor(color);
    buffer->StrokeLine(BPoint(X0, Y0), BPoint(X1, Y1));
    buffer->Window()->Unlock();
}


void
W_MaskText(W_Window window, int x, int y, W_Color color, char *str, int len, W_Font font)
{
    BView *buffer = (BView *)window->buffer;
    BFont *f = (BFont *)font;

    struct font_height fh;
    f->GetHeight(&fh);

    buffer->Window()->Lock();
    buffer->SetHighColor(color);
    buffer->SetFont(f);
    buffer->SetDrawingMode(B_OP_COPY);
    buffer->DrawString(str, BPoint(x, y + 0.999 + fh.ascent));
    buffer->Window()->Unlock();
}


void
W_NextEvent(W_Window window, W_Event *wevent)
{
    GalagaView *view = (GalagaView *)window->view;
    view->nextEvent(wevent);
}


int
W_StringWidth(char *string, W_Font font)
{
    return int(0.999 + ((BFont *)font)->StringWidth(string));
}


void
W_UngrabPointer(W_Window window)
{
    BView *view = (BView *)window->view;
    view->Window()->Lock();
    view->SetEventMask(0);
    view->Window()->Unlock();
}


void
beos_obscure_cursor()
{
    be_app->ObscureCursor();
}


void
beos_cycle_color(W_Color *color)
{
    static int delta = 0x10;

    int newValue = color->red + delta;
    if ((delta > 0 && newValue > 0xff) || (delta < 0 && newValue < 0x80)) {
	delta = -delta;
    }

    newValue = color->red + delta;
    color->red = color->green = color->blue = newValue;
}


void
beos_free_image(W_Image *image)
{
    BBitmap *bitmap = (BBitmap *)image->pixmap;
    if (bitmap != NULL) {
	delete bitmap;
	image->pixmap = NULL;
    }
}


void
beos_free_fonts()
{
    BFont *f = (BFont *)W_BigFont;
    if (f != NULL) {
	delete f;
	W_BigFont = NULL;
    }

    f = (BFont *)W_RegularFont;
    if (f != NULL) {
	delete f;
	W_RegularFont = NULL;
    }
}

